#include <S:\Engine\Graphics\Graphics2D.h>

#include <S:\Engine\Log\Logger.h>

#include <string>

using namespace Microsoft::WRL;

#pragma comment (lib, "d3d11.lib")
#pragma comment (lib, "d2d1.lib")
#pragma comment (lib, "dwrite.lib")

D2DGraphics::D2DGraphics() {
}
D2DGraphics::~D2DGraphics() {
	Log_Write("** Releasing of 2D Devices **");
	Destroy();
	DestroyAllWindows();
	DeleteGUIContext();
	Log_Write("** 2D Devices Released **");
}

bool D2DGraphics::InitD2D(ID3D12Device *pDevice, ComPtr<ID3D12CommandQueue> pCommandQueue, ComPtr<ID3D12Resource> pRenderTargets[]) {
	UINT d3d11DeviceFlags = D3D11_CREATE_DEVICE_BGRA_SUPPORT;

	ErrorHR(D3D11On12CreateDevice(pDevice, d3d11DeviceFlags, nullptr, 0, reinterpret_cast<IUnknown**>(pCommandQueue.GetAddressOf()), 1, 0, &d3d11Device, &pD3D11DeviceContext, nullptr));

	ErrorHR(d3d11Device.As(&pD3D11On12Device));

	D2D1_DEVICE_CONTEXT_OPTIONS deviceOptions = D2D1_DEVICE_CONTEXT_OPTIONS_NONE;
	ErrorHR(D2D1CreateFactory(D2D1_FACTORY_TYPE_SINGLE_THREADED, __uuidof(ID2D1Factory3), &d2dFactoryOptions, &pD2DFactory));

	ComPtr<IDXGIDevice> dxgiDevice;
	ErrorHR(pD3D11On12Device.As(&dxgiDevice));

	ErrorHR(pD2DFactory->CreateDevice(dxgiDevice.Get(), &pD2DDevice));

	ErrorHR(pD2DDevice->CreateDeviceContext(deviceOptions, &pD2DDeviceContext));
	
	ErrorHR(DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory), &pDWriteFactory));
	
	ErrorHR(CoInitialize(nullptr));
	
	ErrorHR(CoCreateInstance(CLSID_WICImagingFactory1, NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&m_pWICFactory)));
	
	float dpiX{};
	float dpiY{};
	pD2DFactory->GetDesktopDpi(&dpiX, &dpiY);
	D2D1_BITMAP_PROPERTIES1 bitmapProperties = D2D1::BitmapProperties1(D2D1_BITMAP_OPTIONS_TARGET | D2D1_BITMAP_OPTIONS_CANNOT_DRAW,
		D2D1::PixelFormat(DXGI_FORMAT_UNKNOWN, D2D1_ALPHA_MODE_PREMULTIPLIED),
		dpiX,
		dpiY
	);

	for (UINT n = 0; n < 3; n++) {
		D3D11_RESOURCE_FLAGS d3d11Flags = { D3D11_BIND_RENDER_TARGET };
		ErrorHR(pD3D11On12Device->CreateWrappedResource(
			pRenderTargets[n].Get(),
			&d3d11Flags,
			D3D12_RESOURCE_STATE_RENDER_TARGET,
			D3D12_RESOURCE_STATE_PRESENT,
			IID_PPV_ARGS(&pWrappedBackBuffers[n])
		));
		
		ComPtr<IDXGISurface> surface;
		ErrorHR(pWrappedBackBuffers[n].As(&surface));
		
		ErrorHR(pD2DDeviceContext->CreateBitmapFromDxgiSurface(surface.Get(), &bitmapProperties, &pD2DRenderTargets[n]));	
	}

	SetGUIContext(pD2DDeviceContext, pDWriteFactory);
		
	return true;
}

void D2DGraphics::RenderD2D(int FrameIndex) {
	pD3D11On12Device->AcquireWrappedResources(pWrappedBackBuffers[FrameIndex].GetAddressOf(), 1);

	pD2DDeviceContext->SetTarget(pD2DRenderTargets[FrameIndex].Get());
	pD2DDeviceContext->BeginDraw();
	pD2DDeviceContext->SetTransform(D2D1::Matrix3x2F::Identity());

	// UI Render

	RenderUI();

	// -- UI Render

	ErrorHR(pD2DDeviceContext->EndDraw());
	
	pD3D11On12Device->ReleaseWrappedResources(pWrappedBackBuffers[FrameIndex].GetAddressOf(), 1);

	pD3D11DeviceContext->Flush();
}

void D2DGraphics::Destroy() {
	m_pWICFactory.Reset();
	Log_Write("\tWIC Factory Released");
	pD2DFactory.Reset();
	Log_Write("\tD2D Factory Released");
	pD2DDevice.Reset();
	Log_Write("\tDevice 2D Released");
	pDWriteFactory.Reset();
	Log_Write("\tDirectWrite Released");
	d3d11Device.Reset();
	Log_Write("\tDevice 11 Released");

	for (int i = 0; i < 3; ++i) {
		pD2DRenderTargets[i].Reset();
		Log_Write("\tRTarged 2D Released");
		pWrappedBackBuffers[i].Reset();
		Log_Write("\tWrappedBackBuffers Released");
	};

	pD2DDeviceContext.Reset();
	Log_Write("\tDevice 2D Context Released");
	pD3D11On12Device.Reset();
	Log_Write("\tDevice 11On12 Released");
	pD3D11DeviceContext.Reset();
	Log_Write("\tDevice 11 Context Released");
}